#
# (C) Eithery Lab, 2023
# ETL top-level package primitives and service functions
#
import os
from dotenv import load_dotenv
from etl.config.app import ENVIRONMENT_ENV_VAR, DEFAULT_ENVIRONMENT

load_dotenv()


def current_environment() -> str:
    env = os.getenv(ENVIRONMENT_ENV_VAR)
    return env.lower() if env else DEFAULT_ENVIRONMENT



#
# (C) Eithery Lab, 2023
# ETL paths module
# Contains the set of path constants
#
from pathlib import Path

ROOT_DIR: Path = Path(__file__).parent.parent.parent
CONFIG_DIR: Path = ROOT_DIR.joinpath("config")
TEMPLATES_DIR: Path = CONFIG_DIR.joinpath("templates")
TEST_DIR: Path = ROOT_DIR.joinpath("tests")



class FileParsingResult:
    def __init__(self, file_name: str):
        self.file_name = file_name


    @property
    def has_errors(self) -> bool:
        return False


    @property
    def errors_count(self) -> int:
        return 0


    @property
    def has_warnings(self) -> bool:
        return False


    @property
    def warnings_count(self) -> int:
        return 0



#
# (C) Eithery Lab, 2023
# ETL pipeline module
# Implements a pipeline for ETL process
#
import etl
import etl.cli as cli
from typing import Optional
from etl.config.app import AppConfiguration
from etl.std.result import Result, Ok
from etl.models.file_parsing_result import FileParsingResult


def run(
    files: tuple[str, ...],
    template: str,
    config_dir: Optional[str] = None,
    verbose: bool = False
) -> Result[None]:
    if verbose:
        cli.echo(f'Current environment: {etl.current_environment()}')
        cli.echo('Load configuration:')
    config = AppConfiguration.load(config_dir, verbose)
    result = parse_file(template, config)
    if result:
        if result.has_errors:
            cli.warning(f"File '{result.file_name}' has been loaded with errors", prefix = None, prepend_line = True)
            cli.warning(_build_stats_message(result), prefix = None)
        else:
            cli.success(f"File '{result.file_name}' has been loaded successfully")
            if result.has_warnings:
                cli.warning(f'Total {result.warnings_count} warnings detected', prefix = None)
    return Ok()


def parse_file(template: str, config: AppConfiguration) -> FileParsingResult:
    return FileParsingResult('file_name')


# private

def _build_stats_message(result: FileParsingResult) -> str:
    message = f'Total {result.errors_count} errors'
    if result.has_warnings:
        message += f' and {result.warnings_count} warnings'
    return message + ' detected'



#
# (C) Eithery Lab, 2023
# Database configuration module
# Contains database related configuration settings
#
from __future__ import annotations
import os
from typing import Optional
from etl.config.auth_type import AuthType
from etl.config.settings import DbConfigSettings


DEFAULT_DB_DIALECT = 'mssql'
DEFAULT_ODBC_DRIVER = 'ODBC Driver 17 for SQL Server'
DEFAULT_DB_HOST = 'localhost'
DEFAULT_DB_NAME = 'etl_db'
_SSPI_AUTH = ['sspi', 'windows']

DB_HOST_ENV_VAR = 'ETL_DB_HOST'
DB_INSTANCE_ENV_VAR = 'ETL_DB_INSTANCE'
DB_NAME_ENV_VAR = 'ETL_DB_NAME'
DB_USER_ENV_VAR = 'ETL_DB_USER'
DB_PWD_ENV_VAR = 'ETL_DB_PWD'               # nosec
DB_AUTH_TYPE_ENV_VAR = 'DB_AUTH_TYPE'


class DbConfiguration:
    def __init__(
        self,
        dialect: Optional[str] = None,
        driver: Optional[str] = None,
        host: Optional[str] = None,
        instance_name: str = '',
        db_name: Optional[str] = None,
        connection_type: Optional[str] = None,
        uid: Optional[str] = None,
        pwd: Optional[str] = None
    ):
        self._dialect = dialect or DEFAULT_DB_DIALECT
        self._driver = driver or DEFAULT_ODBC_DRIVER
        self._host = host or DEFAULT_DB_HOST
        self._instance_name = instance_name or ''
        self._db_name = db_name or DEFAULT_DB_NAME
        self._connection_type = connection_type or 'default'
        self._uid = uid
        self._pwd = pwd


    @property
    def host(self) -> str:
        return f'{self._host}\\{self._instance_name}' if self._instance_name else self._host


    @property
    def db_name(self) -> str:
        return self._db_name


    @property
    def auth_type(self) -> AuthType:
        return AuthType.SSPI if self._connection_type in _SSPI_AUTH else AuthType.DEFAULT


    @property
    def uid(self) -> Optional[str]:
        return self._uid


    @property
    def pwd(self) -> Optional[str]:
        return self._pwd


    @property
    def connection_string(self) -> str:
        prefix = '' if self.auth_type == AuthType.SSPI else f'{self._uid}:{self._pwd}@'
        return f'mssql+pyodbc://{prefix}{self.host}/{self.db_name}?driver={self._driver}'


    def merge(self, params: Optional[DbConfigSettings]) -> DbConfiguration:
        return DbConfiguration(
            dialect = params.get('dialect', self._dialect),
            driver = params.get('driver', self._driver),
            host = params.get('host', self._host),
            instance_name = params.get('instance_name', self._instance_name),
            db_name = params.get('db_name', self._db_name),
            connection_type = params.get('connection', self._connection_type),
            uid = self._uid,
            pwd = self._pwd
        ) if params else self


    def apply_env_vars(self) -> DbConfiguration:
        return DbConfiguration(
            dialect = self._dialect,
            driver = self._driver,
            host = os.getenv(DB_HOST_ENV_VAR, self._host),
            instance_name = os.getenv(DB_INSTANCE_ENV_VAR, self._instance_name),
            db_name = os.getenv(DB_NAME_ENV_VAR, self._db_name),
            connection_type = os.getenv(DB_AUTH_TYPE_ENV_VAR, self._connection_type),
            uid = os.getenv(DB_USER_ENV_VAR, self._uid),
            pwd = os.getenv(DB_PWD_ENV_VAR, self._pwd)
        )



#
# (C) Eithery Lab, 2023
# Application configuration module
# Contains application level configuration settings
#
from __future__ import annotations
import os
import etl
import etl.config.yaml as yaml
from dotenv import load_dotenv
from pathlib import Path
from typing import Optional
from etl.config.database import DbConfiguration
from etl.config.settings import AppConfigSettings
from etl.paths import CONFIG_DIR, ROOT_DIR, TEMPLATES_DIR


CONFIG_FILE_NAME = '.etlrc.yml'
DEV_CONFIG_FILE_NAME = '.etlrc.dev.yml'
TEST_CONFIG_FILE_NAME = '.etlrc.test.yml'
PROD_CONFIG_FILE_NAME = '.etlrc.prod.yml'

DEV_ENV_NAMES = ['dev', 'development']
TEST_ENV_NAMES = ['test', 'testing']
PROD_ENV_NAMES = ['prod', 'production']

DEFAULT_INBOX_HOME = './data/inbox'
DEFAULT_ENVIRONMENT = 'prod'

ENVIRONMENT_ENV_VAR = 'ETL_ENVIRONMENT'
CONFIG_HOME_ENV_VAR = 'ETL_CONFIG_HOME'


class AppConfiguration:
    def __init__(
        self,
        db: Optional[DbConfiguration] = None,
        inbox: Optional[list[str]] = None,
        templates: Optional[list[str]] = None
    ):
        self._db = db or DbConfiguration()
        self._inbox = inbox or []
        self._templates = templates or []


    @property
    def db(self) -> DbConfiguration:
        return self._db


    @property
    def inbox(self) -> list[str]:
        return self._inbox or [DEFAULT_INBOX_HOME]


    @property
    def templates(self) -> list[Path]:
        return [self._to_absolute_path(template) for template in self._templates] or [TEMPLATES_DIR]


    @staticmethod
    def load(config_option: Optional[str] = None, verbose: bool = False) -> AppConfiguration:
        load_dotenv()
        return AppConfiguration()\
            ._load_from_home(verbose)\
            ._load_from_dir(CONFIG_DIR, verbose)\
            ._load_env_config(CONFIG_DIR, verbose)\
            ._load_from_dir(Path('.'), verbose)\
            ._load_from_env(CONFIG_HOME_ENV_VAR, verbose)\
            ._load_from_option(config_option, verbose)\
            ._apply_env_vars()


    def _load_from_home(self, verbose: bool) -> AppConfiguration:
        return self._load(Path.home(), verbose = verbose)


    def _load_from_dir(self, config_path: Path, verbose: bool) -> AppConfiguration:
        return self._load(config_path, verbose = verbose) if config_path else self


    def _load_from_env(self, env_var: str, verbose: bool) -> AppConfiguration:
        config_dir = os.getenv(env_var)
        return self._load(Path(config_dir), verbose = verbose) if config_dir else self


    def _load_from_option(self, option: Optional[str], verbose: bool) -> AppConfiguration:
        return self._load(Path(option), verbose = verbose) if option else self


    def _load_env_config(self, config_path: Path, verbose: bool) -> AppConfiguration:
        env = etl.current_environment()
        if env in DEV_ENV_NAMES:
            return self._load(config_path, DEV_CONFIG_FILE_NAME, verbose)
        if env in TEST_ENV_NAMES:
            return self._load(config_path, TEST_CONFIG_FILE_NAME, verbose)
        if env in PROD_ENV_NAMES:
            return self._load(config_path, PROD_CONFIG_FILE_NAME, verbose)
        return self


    def _load(
        self,
        config_dir: Path,
        config_file_name: str = CONFIG_FILE_NAME,
        verbose: bool = False
    ) -> AppConfiguration:
        config_path = config_dir.joinpath(config_file_name)
        return yaml.load(config_path, verbose).match(
            ok = self._merge_config,
            err = lambda _: self
        )


    def _apply_env_vars(self) -> AppConfiguration:
        return AppConfiguration(
            db = self.db.apply_env_vars(),
            inbox = self._inbox,
            templates = self._templates
        )


    def _merge_config(self, params: AppConfigSettings) -> AppConfiguration:
        return AppConfiguration(
            db = self.db.merge(params.get('database')),
            inbox = self._prepend_list(self._inbox, params.get('inbox', [])),
            templates = self._prepend_list(self._templates, params.get('templates', []))
        )


    def _prepend_list(self, source: list[str], values: list[str]) -> list[str]:
        return values + source if values else source


    def _to_absolute_path(self, template: str) -> Path:
        path = Path(template)
        return path if path.is_absolute() else ROOT_DIR.joinpath(path)



#
# (C) Eithery Lab, 2023
# Defines CLI application class
# Overrides CLI error handling and help formatting
#
# mypy: allow_subclassing_any
#
import click
import sys
import etl.cli as cli
from typing import Any


class App(click.Group):
    group_class = type

    def main(self, *args: Any, **kwargs: Any) -> None:
        try:
            super().main(*args, standalone_mode = False, **kwargs)
        except click.ClickException as e:
            cli.error(str(e))
            sys.exit(e.exit_code)
        except click.exceptions.Abort:
            cli.error('Aborted!', prefix = None)
            sys.exit(1)


    def format_help(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        click.secho(f'{cli.APP_DISPLAY_NAME}, version {cli.__version__}\n', fg = 'bright_cyan')
        super().format_help(ctx, formatter)



#
# (C) Eithery Lab, 2023
# Settings module
# Contains TypedDict definitions for application and DB configuration settings
#
from typing import TypedDict


class DbConfigSettings(TypedDict, total = False):
    dialect: str
    driver: str
    host: str
    instance_name: str
    db_name: str
    connection: str


class AppConfigSettings(TypedDict, total = False):
    database: DbConfigSettings
    inbox: list[str]
    templates: list[str]



#
# (C) Eithery Lab, 2023
# Error results module
# Contains the set of common error results
#
from pathlib import Path
from etl.std import T
from etl.std.error import Error
from etl.std.result import Result, Err


def FileDoesNotExistError(file_path: Path) -> Result[T]:
    return Err(Error(f"The file '{file_path}' doesn't exist"))



#
# (C) Eithery Lab, 2023
# YAML configuration module
# Supports reading configuration from YAML files
#
import yaml
import etl.cli as cli
from pathlib import Path
from etl.config.settings import AppConfigSettings
from etl.std.error_results import FileDoesNotExistError
from etl.std.result import Result, Ok


def load(file_path: Path, verbose: bool = False) -> Result[AppConfigSettings]:
    if file_path.is_file():
        with open(file_path) as infile:
            settings: AppConfigSettings = yaml.safe_load(infile)
            if verbose:
                cli.echo(f"'{file_path.absolute()}' LOADED")
            return Ok(settings)

    return FileDoesNotExistError(file_path)



#
# (C) Eithery Lab, 2023
# CLI package level primitives
# Provides the app version and service functions for stdout/stderr colorful output
#
import click
import importlib.metadata
from typing import Optional


_ERROR_COLOR = 'red'
_WARNING_COLOR = 'yellow'
_SUCCESS_COLOR = 'green'

APP_DISPLAY_NAME = 'ETL CLI Toolkit'

__version__ = importlib.metadata.version('etl-toolkit')


def error(message: str, prefix: Optional[str] = 'Error:', prepend_line: bool = True) -> None:
    _display_message(message, prefix, color = _ERROR_COLOR, prepend_line = prepend_line)


def warning(message: str, prefix: Optional[str] = 'Warning:', prepend_line: bool = False) -> None:
    _display_message(message, prefix, color = _WARNING_COLOR, prepend_line = prepend_line)


def success(message: str, prefix: Optional[str] = None, prepend_line: bool = True) -> None:
    _display_message(message, prefix, color = _SUCCESS_COLOR, prepend_line = prepend_line)


def echo(message: str) -> None:
    click.echo(message)


# private

def _display_message(message: str, prefix: Optional[str], color: str, prepend_line: bool = False) -> None:
    if prepend_line:
        click.echo()
    combined_message = ' '.join(filter(None, [prefix, message]))
    click.secho(combined_message, fg = color)



#
# (C) Eithery Lab, 2023
# The main CLI application entry point
# Defines available CLI application commands and options
#
# mypy: allow-untyped-decorators
#
import click
import etl.cli as cli
import etl.pipeline
from etl.cli.app import App


@click.group(cls = App)
@click.pass_context
@click.version_option(cli.__version__, '-V', '--version', prog_name = cli.APP_DISPLAY_NAME)
@click.help_option('-h', '--help')
def main(ctx: click.Context) -> None:
    ctx.ensure_object(dict)


@main.command(name = 'load', help = 'Load files belong to the given template.')
@click.argument('files', nargs = -1, required = False)
@click.option('-t', '--template', metavar = 'TEMPLATE_NAME', help = 'Template applied to the file(s) being loaded.')
@click.option('-c', '--config', metavar = 'CONFIG_PATH', help = 'Path to the custom app configuration file.')
@click.option('-v', '--verbose', type = bool, is_flag = True, default = False,
    help = 'Verbose mode. Display additional details.')
@click.help_option('-h', '--help')
def load_files(files: tuple[str, ...], template: str, config: str, verbose: bool) -> None:
    print('Files:', files)
    print('Template', template)
    etl.pipeline.run(files, template, config, verbose)



#
# (C) Eithery Lab, 2023
# Error class
# Represents a basic value for error result
#
from __future__ import annotations
from typing import Optional


class Error:
    def __init__(
        self,
        message: str,
        code: Optional[str] = None,
        description: Optional[str] = None,
        inner_error: Optional[Error] | Optional[BaseException] = None
    ):
        self._message = message
        self._code = code
        self._description = description
        self._inner_error = inner_error


    @property
    def code(self) -> Optional[str]:
        return self._code


    @property
    def message(self) -> str:
        return self._message


    @property
    def description(self) -> Optional[str]:
        return self._description


    def __repr__(self) -> str:
        return self._message


def error(message: str) -> Error:
    return Error(message)



#
# (C) Eithery Lab, 2023
# Result[T, E] class
# Represents the result of operation with two possible outcomes
# Result[T, E] = Ok(T) | Err(E)
# It contains either a success result OK(t) where t is a wrapped returned value of type T
# or an error result Err(e) where e is an error of type E explaining what went wrong
#
from __future__ import annotations
from etl.std import T, R
from typing import Generic, Optional, Callable, cast
from etl.std.error import Error, error as to_error
from etl.std.exceptions import InvalidResultException


class Result(Generic[T]):
    def __init__(self, value: Optional[T] = None, error: Error | str | None = None):
        self._value = value
        self._error = _convert_to_error(error)
        if isinstance(value, Error):
            self._value = None
            self._error = value


    @property
    def is_ok(self) -> bool:
        return self._error is None


    @property
    def is_error(self) -> bool:
        return not self.is_ok


    def unwrap(self) -> T:
        return self.expect("Cannot unwrap a value for the Error result")


    def expect(self, message: str) -> T:
        if self.is_ok:
            return cast(T, self._value)
        raise InvalidResultException(message)


    def ok(self) -> Optional[T]:
        return self._value if self.is_ok else None


    def err(self) -> Optional[Error]:
        return self._error if self.is_error else None


    def match(self, ok: Callable[[T], R], err: Callable[[Error], R]) -> R:
        return ok(cast(T, self._value)) if self.is_ok else err(cast(Error, self._error))


    def map(self, func: Callable[[T], R]) -> Result[R]:
        return self.match(
            ok = lambda v: Ok(func(v)),
            err = lambda e: Err(e)
        )


    def bind(self, func: Callable[[T], Result[R]]) -> Result[R]:
        return self.match(
            ok = lambda v: func(v),
            err = lambda e: Err(e)
        )


    def __repr__(self) -> str:
        return self.match(lambda v: f'Ok({v})', lambda e: f'Err({e})')




def Ok(value: Optional[T] = None) -> Result[T]:
    return Result(value)


def Err(error: Error) -> Result[T]:
    return Result(error = error)


# private

def _convert_to_error(error: Error | str | None) -> Optional[Error]:
    err = to_error(error) if error and isinstance(error, str) else error
    return cast(Optional[Error], err)
